from os import chdir

import click
from pathlib import Path

from .req import Reqs
from .utils import activate_venv_and_run, get_python_command, get_venv_path


def venv_path_option(func):
    """A decorator for common and same options for venv path"""
    func = click.option(
        "-v", "--venv-path",
        type=click.Path(exists=True, file_okay=False, dir_okay=True),
        help="Path to the venv directory"
    )(func)
    return func


@click.group()
def main():
    pass


@main.command()
@venv_path_option
@click.option(
    "-a", "--arguments",
    type=click.STRING,
    help="Specify additional arguments to pass to the Python file during execution. "
         "Use quotes for multiple arguments or arguments containing spaces. "
         "For example: --arguments=\"-a 42 --verbose\""
)
@click.argument(
    "file_path",
    type=click.Path(exists=True, file_okay=True, dir_okay=False)
)
def file(file_path: str, venv_path: str | None, arguments: str | None):
    """Execute Python file"""

    if not file_path.endswith(".py"):
        raise click.BadParameter("File extension must be .py")

    current_dir = Path.cwd()  # Get the current directory
    file_path = (current_dir / file_path).resolve()
    file_dir = file_path.parent  # Get the file directory
    venv_path = get_venv_path(venv_path, file_dir)

    # Activate venv and run Python file
    activate_venv_and_run(
        f"{get_python_command()} \"{file_path}\" {arguments if arguments is not None else ''}",
        venv_path,
        file_dir
    )


@main.command()
@venv_path_option
@click.argument("command")
def cmd(command: str, venv_path: str | None):
    """Execute a shell command"""

    venv_path = get_venv_path(venv_path)
    activate_venv_and_run(command, venv_path)


@main.command()
@venv_path_option
@click.option(
    "-w", "--workdir",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help="Change the working directory before executing commands"
)
def shell(venv_path: str | None, workdir: str | None):
    """Open shell to execute commands (To exit shell, enter "exit")"""

    current_dir = Path.cwd()  # Get the current working directory
    if workdir is not None:
        chdir(workdir)  # Change the working directory
        working_dir = Path(workdir)
        venv_path = get_venv_path(venv_path, working_dir)
    else:
        venv_path = get_venv_path(venv_path)

    while True:
        command = input("> ")
        if command in "exit":
            break
        activate_venv_and_run(command, venv_path)


@main.command()
@venv_path_option
@click.option(
    "-p", "--project", type=click.Path(exists=True, file_okay=False, dir_okay=True),
    default='.', show_default=True, help="Path to project for finding requirements"
)
@click.option(
    "-s", "--standard",
    type=click.Choice(['true', 'false']), help="Filter based on Python's built-in modules."
)
@click.option(
    "-e", "--exist",
    type=click.Choice(['true', 'false']), help="Filter based on modules installed in venv."
)
@click.option('--version', is_flag=True, help='Display module versions.')
@click.option(
    "-o", "--output",
    type=click.Path(file_okay=True, dir_okay=False), default=None, flag_value="requirements.txt",
    help="Save results to a file. If provided without a value, defaults to 'requirements.txt'"
)
@click.option("-i", "--install", is_flag=True, help="Install the found packages in venv")
def reqs(project, exist, standard, version, output, venv_path, install):
    """Find requirements for a project or install them after finding"""

    project_path = Path(project)
    venv_path = get_venv_path(venv_path, project_path)

    # Check only for third-party modules
    if exist or install:
        if exist is None:
            version = False
            exist = 'false'
        standard = 'false'

    if version:
        click.echo(click.style("Warning: It may take some time to display the results because "
                               "need to search PyPi to find the version of some modules.\n", fg="yellow"))

    requirements = Reqs(project_path, exist, standard, venv_path, version).find()

    if requirements:
        result = '\n'.join(
            f"{name}=={version}" if version else name
            for name, version in requirements.items()
        )

        if output:
            # Save to a file
            with open(output, 'w', encoding='utf-8') as f:
                f.write(result + '\n')
            click.echo(f'Requirements saved in "{output}"')
        elif not install:
            click.echo(result)

        # Install requirements
        if install:
            display_venv_path = 'System' if venv_path is None else str(venv_path)
            click.echo(click.style('Venv path: ', fg='blue') + display_venv_path)
            click.echo(click.style('Packages: ', fg='blue') + ' - '.join(requirements.keys()))

            entry = input('\nInstall? (yes/no) ')
            if entry in ('yes', 'y'):
                for r in requirements.keys():
                    activate_venv_and_run(
                        f'{get_python_command()} -m pip install {r}',
                        venv_path
                    )
    else:
        click.echo(click.style('No requirements found.', fg='red'))


if __name__ == "__main__":
    main()



import os
import sys
import click
import logging
import subprocess
from pathlib import Path


class ExitOnErrorHandler(logging.StreamHandler):
    def emit(self, record):
        super().emit(record)  # Display error message
        if record.levelno >= logging.ERROR:  # If the message level was ERROR or higher
            sys.exit(1)  # Exit the program with error code


logger = logging.getLogger(__name__)

# Configure logging only if not already configured
if not logger.hasHandlers():
    handler = ExitOnErrorHandler(sys.stderr)  # Use a custom handler
    formatter = logging.Formatter("{levelname}: {message}", style="{")
    handler.setFormatter(formatter)

    logger.setLevel(logging.ERROR)  # Default logging level
    logger.addHandler(handler)


def get_venv_path(venv_path: str | None, dir_path: Path|None = None) -> Path | None:
    """Specify venv path"""

    current_dir = Path.cwd()
    if venv_path is None:
        if dir_path and (dir_path / "venv").exists():
            venv_path = dir_path / "venv"
        elif (current_dir / "venv").exists():
            venv_path = current_dir / "venv"

        if venv_path is not None:
            click.echo(click.style("venv detected: ", fg="green") + str(venv_path.resolve()) + "\n")
    else:
        venv_path = (current_dir / venv_path).resolve()

    return venv_path


def get_python_command() -> str:
    """Check for python3 (for Mac and Linux)"""

    # Is there python3?
    try:
        subprocess.run("python3 --version", check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        return "python3"
    except Exception:
        pass

    # If python3 is not available, check python
    try:
        result = subprocess.run("python --version", check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                shell=True)
        version = result.stdout.decode().strip() or result.stderr.decode().strip()
        if version.startswith("Python 3"):
            return "python"
    except Exception:
        pass

    logger.error("No compatible Python 3 interpreter found on the system.")


def activate_venv_and_run(command: str, venv_path: Path | None = None, chdir_path: Path | None = None,
                          capture_output: bool = False) -> str|None:
    """
    Activate venv and run command
    :param venv_path: Path to the venv directory
    :param command: Command to be executed
    :param chdir_path: The directory path to which the current path should be changed
    :param capture_output: Should the function return the result of the command or display it in the terminal?
    """

    try:
        # Save the current working directory (where 'pav' is executed)
        initial_cwd = Path.cwd()

        # Change current working directory to the script's directory
        if chdir_path is not None:
            os.chdir(chdir_path)

        if venv_path is not None:
            venv_path = (initial_cwd / venv_path).resolve()

            # venv activation script path
            if os.name == "nt":  # Windows
                activate_script = venv_path / "Scripts" / "activate"
            else:  # Mac/Linux
                activate_script = venv_path / "bin" / "activate"

            if not activate_script.exists():
                logger.error(f"Cannot find activation script at {activate_script}")

            # Create a command to run the activation script and execute the command
            if os.name == "nt":  # Windows
                cmd = f'"{activate_script}" & {command}'
                output = subprocess.run(cmd, shell=True, capture_output=capture_output, text=True)
            else:  # Mac/Linux
                cmd = f'source "{activate_script}" && {command}'
                # To use the "source" command, must change the shell from "/bin/sh" to "/bin/bash"
                output = subprocess.run(cmd, shell=True, executable="/bin/bash", capture_output=capture_output, text=True)
        else:
            # If no venv_path is provided, run the command directly
            output = subprocess.run(command, shell=True, capture_output=capture_output, text=True)

        return output.stdout.strip() if capture_output else None
    except Exception as e:
        logger.error(e)



import re
import requests
from pathlib import Path
from sysconfig import get_path
from importlib.util import find_spec
from .utils import activate_venv_and_run, get_python_command

# Directories that aren't suitable for searching requirements
EXCLUDED_DIRS = (
    "venv", ".venv", "__pycache__", ".git", ".hg", ".svn", ".idea", ".vscode",
    "node_modules", "dist", "build", "migrations", "logs", "coverage", ".coverage",
    "staticfiles", "media", ".pytest_cache"
)


def is_relative_to(path):
    """Check if path includes 'EXCLUDED_DIRS'"""

    # Create a regex pattern to search for exclude dirs
    patterns = map(re.escape, EXCLUDED_DIRS)
    final_pattern = '|'.join(rf'\\?{d}\\' for d in patterns)

    return bool(re.search(final_pattern, str(path)))


def is_standard_library(module_name: str) -> bool:
    """Check whether a module belongs to the Python standard library"""
    spec = find_spec(module_name)
    if not spec or not spec.origin:
        return False  # The module was not found, so it is not standard

    return "site-packages" not in spec.origin and "dist-packages" not in spec.origin


def get_pypi_names(modules: list[str]) -> dict:
    """Get PyPI module names from mapping file"""
    mapping_path = Path(__file__).parent / "mapping"
    with open(mapping_path, "r") as f:
        names = dict(line.strip().split(":") for line in f)
    return {p: names.get(p, p) for p in modules}


class Reqs:
    def __init__(self, project: Path, exist: str|None, standard: str|None, venv_path: Path|None, need_version: bool):
        self.project = project
        self.exist = exist
        self.standard = standard
        self.venv_path = venv_path
        self.version = need_version

    def is_internal_module(self, module_name: str, p_resolved: Path) -> bool:
        """
        Check whether a module is part of the project or an external library
        If the module is in the project path, it is internal
        """
        p_parent = p_resolved.parent
        module_path_parent = (p_parent / (module_name.replace(".", "/") + ".py")).resolve()
        module_dir_parent = (p_parent / module_name.split(".")[0]).resolve()

        module_path = (self.project / (module_name.replace(".", "/") + ".py")).resolve()
        module_dir = (self.project / module_name.split(".")[0]).resolve()

        # If a file or directory associated with this module exists, it is internal
        return module_path.exists() or module_dir.exists() or module_dir_parent.exists() or module_path_parent.exists()

    def get_site_packages(self) -> Path:
        """Find the correct site-packages path inside a virtual environment"""
        site_packages_relative = get_path("purelib", vars={"base": self.venv_path})
        return Path(site_packages_relative)

    def is_module_exist(self, module_name: str) -> bool:
        """Check if module is installed inside venv"""
        if self.venv_path is not None:
            site_packages = self.get_site_packages()
            spec = (site_packages / module_name).exists() or (site_packages / f'{module_name}.py').exists()
        else:
            spec = bool(find_spec(module_name))
        return spec

    def conditions(self, module_name: str) -> bool:
        result = set()

        # Found on the system (i.e. installed)
        if self.exist:
            spec = self.is_module_exist(module_name)
            result.add(spec if self.exist == 'true' else not spec)

        # Filter based on built-in Python module
        if self.standard:
            spec = is_standard_library(module_name)
            result.add(spec if self.standard == 'true' else not spec)

        return all(result)

    def get_module_version(self, module_name: tuple) -> str|None:
        """Get module version (If the module is not installed, its version will be taken from PyPi.)"""
        if not is_standard_library(module_name[0]):
            if self.is_module_exist(module_name[0]):
                # Get module info from pip
                output = activate_venv_and_run(
                    f"{get_python_command()} -m pip show {module_name[1]}",
                    self.venv_path,
                    capture_output = True
                ).lower()
                # Return module version if it exists
                return next(
                    (line.split(":", 1)[1].strip()
                     for line in output.splitlines()
                     if line.startswith("version:")),
                    None  # default if not found
                )
            else:
                # Get version from PyPi
                url = f"https://pypi.org/pypi/{module_name[1]}/json"
                response = requests.get(url)
                if response.status_code == 200:
                    data = response.json()
                    return data["info"]["version"]

    def find(self) -> dict:
        """Find all requirements for a project and return a list of them"""
        module_names = set()

        # Read python files to find import module names
        for p in self.project.rglob('*.py'):
            p_resolved = p.resolve()

            # Filter excluded paths
            if is_relative_to(p_resolved):
                continue

            # Read file line by line
            with open(p_resolved, 'r', encoding='utf-8', errors='ignore') as f:
                for line in f:
                    line = line.strip()

                    # Find lines that import something
                    if line.startswith(('import ', 'from ')):
                        parts = line.split()[1]
                        module_name = parts.split('.')[0]  # Get the original module name

                        if not self.is_internal_module(parts, p_resolved) and self.conditions(module_name):
                            module_names.add(module_name)

        pypi_names = get_pypi_names(sorted(module_names))
        requirements = {m[1]: self.get_module_version(m) if self.version else None for m in pypi_names.items()}
        return requirements



from setuptools import setup, find_packages


setup(
    name="pav",
    version="1.4.0",
    author="Arian",
    author_email="ariannasiri86@gmail.com",
    description="PAV is a Python library with features to easily work with venv",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/ArianN8610/PAV",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=["click==8.1.8"],
    entry_points={"console_scripts": ["pav=pav.cli:main"]}
)
